/*
 * Copyright 2005,2006 WSO2, Inc. http://www.wso2.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.ws.dataservice.beans;

import java.util.ArrayList;
import java.util.Iterator;

import javax.xml.namespace.QName;

import org.apache.axiom.om.OMElement;

/**
 * Object Model for the data service configuration.
 * <data name="sample">            
 *   <config>..</config>
 *    <operation>..</operation>
 *    <query>..</query>
 * </data>       
 */
public class Data {
	private String name;
	private Config config;
	private ArrayList operations; //Arraylist of Operation objects
	private ArrayList queries; //Arraylist of Query objects	
	private ArrayList resources; //Arraylist of Resource objects
	public String getName() {
		return name;
	}
	public void setName(String name) {
		this.name = name;
	}
	public Config getConfig() {
		return config;
	}
	public void setConfig(Config config) {
		this.config = config;
	}
	public ArrayList getOperations() {
		return operations;
	}
	public ArrayList getQueries() {
		return queries;
	}
	public ArrayList getResources() {
		return resources;
	}
	
	/**
	 * Schema validation for DS configuration file 
	 */
	public void validate(){
		//TODO
	}
	
	/**
	 * populate Object model using OM representation of 
	 * configuration file
	 * 
	 */
	public void populate(OMElement dsXml){
		//populate config objects
		//TODO
		
		//populate query objects
		Iterator queryElements = dsXml.getChildrenWithName(new QName("query"));
		queries = new ArrayList();
		Query query;
		while(queryElements.hasNext()){
			OMElement queryEle = (OMElement)queryElements.next();
			OMElement sqlEle = queryEle.getFirstChildWithName(new QName("sql"));
			
			query = new Query();
			query.setId(queryEle.getAttributeValue(new QName("id")));
			query.setConfigToUse(queryEle.getAttributeValue(new QName("useConfig")));
			
			if(sqlEle != null){
				query.setSql(sqlEle.getText());
			}
			Param[] params = getParams(queryEle.getChildrenWithName(new QName("param")));
			query.setParams(params);
			
			//Add to object model
			queries.add(query);
		}

		
		
		//populate operation objects
		Iterator operationElements = dsXml.getChildrenWithName(new QName("operation"));
		operations = new ArrayList();
		Operation operation;
		while(operationElements.hasNext()){
			OMElement operationEle = (OMElement)operationElements.next();
			OMElement callQueryEle = operationEle.getFirstChildWithName(new QName("call-query"));
			
			operation = new Operation();
			operation.setName(operationEle.getAttributeValue(new QName("name")));
			CallQuery callQuery = getCallQuery(callQueryEle);
			operation.setCallQuery(callQuery);
			
			//Add to object model
			operations.add(operation);
		} 
		
		//populate resource objects
		Iterator resourcesElements = dsXml.getChildrenWithName(new QName("resource"));
		resources = new ArrayList();
		Resource resource;
        while(resourcesElements.hasNext()){
        	OMElement resourceEle = (OMElement) resourcesElements.next();
        	OMElement callQueryEle = resourceEle.getFirstChildWithName(new QName("call-query"));
        	
        	resource = new Resource();
        	resource.setMethod(resourceEle.getAttributeValue(new QName("method")));
        	resource.setPath(resourceEle.getAttributeValue(new QName("path")));
        	CallQuery callQuery = getCallQuery(callQueryEle);
        	resource.setCallQuery(callQuery);
        	
        	//Add to object model
        	resources.add(resource);
        }

	}
	
	/**
	 * Populates CallQuery object using it's OM representation.
	 * eg.
	 *   <call-query href="addProduct">
	 *     <with-param name="id" query-param="id" />  
	 *     <with-param name="name" query-param="name" />  
	 *     <with-param name="price" query-param="price" />
	 *   </call-query>  
	 *     
	 * @param callQueryEle
	 * @return CallQuery Object
	 */
	private CallQuery getCallQuery(OMElement callQueryEle){
		CallQuery callQuery = new CallQuery();
		callQuery.setHref(callQueryEle.getAttributeValue(new QName("href")));
		
		//Populate WithParam objects
		ArrayList withParams = new ArrayList();		
		Iterator itrWithParams = callQueryEle.getChildrenWithName(new QName("with-param"));
		while (itrWithParams.hasNext()) {
			OMElement withParamEle = (OMElement) itrWithParams.next();
			WithParam withParam = new WithParam(
					withParamEle.getAttributeValue(new QName("name")),
					withParamEle.getAttributeValue(new QName("query-param"))
					); 
			withParams.add(withParam);
		}
		callQuery.setWithParams(withParams);
		return callQuery;
	}
	
	/**
	 * Returns operation object containing passed operation name
	 * @param operationName
	 * @return
	 */
	public Operation getOperation(String operationName){
		Iterator itrOperations = operations.iterator();
		while(itrOperations.hasNext()){
			Operation operation = (Operation)itrOperations.next();
			if(operation.getName().equals(operationName)){
				return operation;
			}
		}
		return null;
	}
	
	/**
	 * Returns an array of Param objects, when an iterator of param elements is passed in
	 * @param paramItr
	 * @return
	 */
	public Param[] getParams(Iterator paramItr){
		Param param;
		ArrayList paramList = new ArrayList();		
		int ordinal = 0; 
		while (paramItr.hasNext()) {
			ordinal++;
			OMElement paramElement = (OMElement) paramItr.next();

			//start: work-a-round to maintain backward compatibility
			String userSetOrdinalValue = paramElement.getAttributeValue(new QName("ordinal"));
			if(userSetOrdinalValue == null || userSetOrdinalValue.trim().length() == 0){
				userSetOrdinalValue = String.valueOf(ordinal);
			}
			//end: work-a-round to maintain backward compatibility
			
			param = new Param(paramElement.getAttributeValue(new QName("name")),
						paramElement.getAttributeValue(new QName("sqlType")),
						paramElement.getAttributeValue(new QName("type")),
						userSetOrdinalValue);
			paramList.add(param);
		}
		Param[] params = new Param[paramList.size()];
		paramList.toArray(params);
		return params;
	}
	
	/**
	 * Returns query object containing passed query id
	 * @param queryId
	 * @return
	 */
	public Query getQuery(String queryId){
		Iterator itrQueries = queries.iterator();
		while(itrQueries.hasNext()){
			Query query = (Query)itrQueries.next();
			if(query.getId().equals(queryId)){
				return query;
			}
		}
		return null;
	}	
}
